#############################################################################
##
#F  PPart( <G>, <g>, <prime> )...................return <prime> - part of <g>
##
PPart := function( G, g, prime )

    local o, facs, x, p, q, gcd;

    if g = G.identity  then 
        return g; 
    fi;
    o := Order( G, g );
    facs := Factors( o );
    if not prime  in facs  then 
        return G.identity;
    else
        x := Length( Filtered( facs, y -> y = prime ) );
        p := prime ^ x;
        q := o / p;
        if q = 1 then return g; fi;
        gcd := Gcdex( p, q );
        return g ^ ( gcd.coeff2 * q );  
    fi;
end;

#############################################################################
##
#F  FCommutatorSubgroup( <R>, <L>, <K> )...........calculate [<L>, <R>] * <K> 
##
##  <L> / <K> is assumed to be an elementary abelian normal factor of the 
##  parent group.
##
FCommutatorSubgroup := function( R, L, K )

    local  G, gens, Rigs, C, r, g, c;

    G    := Parent( L );
    gens := FactorArg( L, K ).generators;
    Rigs := FactorArg( R, L ).generators;
    C := [];
    for r  in Rigs  do
        for g  in gens  do
            c := Comm( r, g );
            if not c in K then
               AddSet( C, c );
            fi;
        od;
    od;
    return G.operations.MergedIgs( G, K, C, true );
end;
    
#############################################################################
##
#F  IsFCentralFactor( <R>, <L>, <K> ).................check [<L>, <R>] <= <K> 
##
##  <L> / <K> is assumed to be an elementary abelian normal factor of the 
##  parent group.
##
IsFCentralFactor := function( R, L, K )

    local  gensLK, gensR, l, r, c;

    gensLK := FactorArg( L, K ).generators;
    gensR  := R.generators;
    for l  in gensLK  do
        for r  in gensR  do
            c := Comm( l, r );
            if not c in K  then
                return false;
            fi;
        od;
    od;
    return true;
end;

#############################################################################
##
#F  MyExponents( <system>, <u>, <F>, <first>, <next> ) . . . . . . . . .local
##
MyExponents := function( system, u, F, first, next )
    local   depths, expo, d, i, tmp, max;

    # catch trivial case
    if Length( system.base ) = 0 then
        return [];
    fi;

    # set up
    max    := Length( system.base );
    depths := system.depths{[first..max]};
    expo   := List( [first..max], x -> 0 );

    # iterate
    d := DepthAgWord( u );
    i := Position( depths, d );
    while not IsBool(i) do
        tmp := LeadingExponentAgWord( u ) mod RelativeOrderAgWord( u );
        expo[ i ] := tmp;
        u := system.base[first+i-1]^tmp mod u;
        d := DepthAgWord( u );
        i := Position( depths, d );
    od;

    # cut
    expo := expo{[1..next-first+1]};

    # throw it into the right field
    if IsFFE( One( F ))  then
        MakeVecFFE( expo, One( F ) );
        return expo;
    elif IsInt( One( F ) )  then
    	return expo;
    else
        return expo * One( F );
    fi;

end;

#############################################################################
##
#F  IsFormation( <form> ) . . . . . . . . . . . . . . . . . . . . . . . check 
##
IsFormation := function( form )

    if not IsRec( form ) then
        return false;
    fi;
    return IsBound( form.isFormation ) and form.isFormation;
end;

