#############################################################################
##
#F  SagGroupOps.LeastBadFNormalizerIndex( system, R, i ) . . . . . . . .local
##
SagGroupOps.LeastBadFNormalizerIndex := function( system, R, i )

    local bad,       # least bad  index
          w,         # commutator
          exponents, # exponent vector of w  in layer
          max,       # composition length
          gens,      # of R
          g, pg,     # some element with its prime
          k;         # index

    max  := Length( system.base );
    gens := R.generators;

    bad := max + 1;
    for g  in gens  do
        w := Comm( g, system.base[ i ] );

        # get prime
        pg := RelativeOrderAgWord( g );
        if w <> system.H.identity   then
            exponents := MyExponents( system, w, Integers, i+1, max );
            k := 1;

            # run through exponent list until bad entry is found
            while k <= Length( exponents )  do
                if exponents[ k ] <> 0 and 
                   system.weights[ i ][ 3 ] = system.weights[ k+i ][ 3 ]  and 
                   not system.Fcentral[ k+i ] 
                then
                    if k+i < bad  then
                        bad := k+i;
                    fi;
                    k := Length( exponents ) + 1;
                else
                    k := k + 1;
                fi;
            od;
        fi;

        # if bad is minimal return; otherwise go on
        if i = bad - 1  then
            return bad;
        fi;
    od;
    return bad;
end;

#############################################################################
##
#F  SagGroupOps.ChangeGenerator( system, R, i ) . . . . . . . . . . . . local
##
SagGroupOps.ChangeGenerator := function ( system, R, i )

    local max,               # composition length
          k,                 # smallest bad index
          layer,             # layer with bad  index
          first,             # first index of this layer
          next,              # first index of next layer
          size,              # size of layer
          gensNM,            # gens of layer 
          pk, field,         # involved prime and field
          idmat,             # identity matrix
          e,                 # generator of R
          aij,               # operating element
          g,                 # commutator
          A, v,              # one equation system
          E, V,              # simultaneuos linear equation system
          solution,          # one solution of simultaneuos system or false
          l;                 # index

    max := Length( system.base );
    k   := system.H.operations.LeastBadFNormalizerIndex( system, R, i );

    # trivial case
    if k > max  then
        InfoForm2("#I change gens: ", i, " has no bad index\n");
        return;
    fi;
    InfoForm2("#I change gens: ", i, " has bad index = ", k, "\n");

    # get the layer
    layer  := system.Flayers[ k ];
    first  := system.Ffirst[ layer ];
    next   := system.Ffirst[ layer+1 ];
    size   := next - first;
    gensNM := system.base{[first..next-1]};

    # InitializeSystem inhomogenous system  
    V := [ ];
    E := List( [ 1 .. size ], x -> [ ] );

    # get prime and field
    pk    := system.weights[ k ][ 3 ];
    field := GF( pk );
    idmat := IdentityMat( size, field );

    # run through commutators
    for e in R.generators do

        # operation matrix
        aij := e ^ system.base[ i ];
        A := List( gensNM, 
             x -> MyExponents( system, x^aij, field, first, next-1 ) );
        if A <> idmat then
            A := A - idmat;

            # inhom part
            g := Comm( e, system.base[ i ] );
            v := Sublist( MyExponents( system, g, field, i+1, next-1 ),
                          [ first-i..next-1-i] );

            # append to system 
            for l  in [ 1 .. size ]  do
                Append( E[ l ], A[ l ] );
            od;
            Append( V, v );
        fi;
    od;

    # try to solve inhomogenous systems simultaneously
    solution := SolutionMat( E, V );

    # calculate new i-th base element
    g := system.base[i];
    for l  in [ 1..size ]  do
        g := g * gensNM[l] ^ Int(solution[l]);
    od;
    system.base[i] := g;

    # and recur
    system.H.operations.ChangeGenerator( system, R, i );
end;

#############################################################################
##
#F  SagGroupOps.InducedNilpotentSeries( G, U ) . . . .nilpotent series of <U>
##
SagGroupOps.InducedNilpotentSeries := function( G, U )

    local  gens, max,      # cgs of U and length
           series,         # induced nilpotent series
           head,           # run through nilpotent series of G
           found,          # flag
           H,              # new subgroup
           i;              # index

    # check
    if IsBound( U.inducedNilpotentSeries ) then
        return U.inducedNilpotentSeries;
    fi;

    # set up
    gens   := Cgs( U );
    max    := Length( gens );
    i      := 1;
    series := [ Copy( U ) ];

    # loop over nilpotent series of G
    for head  in G.head{ [ 2 .. Length( G.head ) ] }  do

        # get first generator of next subgroup
        found := false;
        while i <= max and Depth( gens[ i ] ) < head  do
            found := true;
            i     := i + 1;
        od;

        # if there is a new subgroup
        if found then
            H := G.operations.AgSubgroup( G, gens{ [ i .. max ] }, false );
            Add( series, H );
        fi;
    od;
    U.inducedNilpotentSeries := series;

    # return result
    return series;
end;

#############################################################################
##
#F  SagGroupOps.PComplement( G, U, p ) . . . . . . . . . . compute G^p \cap U
##
SagGroupOps.PComplement := function( G, U, p )

    local g, gens;

    # calculate Hall subgroup
    gens := [ ];
    Cgs( U );
    for g  in U.cgs  do
        if RelativeOrderAgWord( g ) <> p  then
            Add( gens, g );
        fi;
    od;
    return U.operations.AgSubgroup( U, gens, false );
end;

#############################################################################
##
#F  SagGroupOps.RefinedBaseLayer( G, R, layer ) . . . . . . . . . . . . local
##
SagGroupOps.RefinedBaseLayer := function( G, R, layer )

    local max,                   # length of cgs of G
          first, next, wt,       # of layer
          head, tail, hhead,     # catch part of nilpotent head series 
          firsts, base, central, # output
          cgs,                   # cgs of R
          depths,                # depths of cgs of R
          N, M,                  # actual layer
          U,                     # subgroup
          index,                 # to check the progress
          series, h;             # nilpotent series of Res  

    # check argument
    if IsList( R ) then
        Error("support and local residuals not consistent \n");
    fi;

    # get the layer
    first  := G.first[ layer ];
    next   := G.first[ layer+1 ];
    wt     := G.weights[ first ];

    # set up 
    max    := Length( G.generators );

    # get the nilpotent factor
    head   := G.head[ wt[1] ];
    tail   := G.tail[ wt[1] ];

    # set up result
    firsts := [ ];
    base   := [ ];
    central:= [ ];

    # first check whether we know that R centralises layer
    cgs := Cgs( R );
    if Length( cgs ) = 0 or DepthAgWord( cgs[1] ) >= head or wt[1] = 1  then
        return rec( base    := G.generators{ [first..next-1] },
                    firsts  := [ next ],
                    central := List( [first..next-1], x -> true ) );
    fi;

    # now check whether we know that the layer is R-hypereccetric 
    depths := List( cgs, x -> DepthAgWord( x ) );
    hhead  := G.head[ wt[1]-1 ];
    if wt[2] = 1 and ForAll([hhead..head-1], x -> x in depths) then 
        return rec( base    := G.generators{ [first..next-1] },
                    firsts  := [ next ],
                    central := List( [first..next-1], x -> false ) );
    fi;

    # now start to compute layers
    N := G.operations.AgSubgroup( G, G.generators{[first..max]}, true); 
    M := G.operations.AgSubgroup( G, G.generators{[next..max]}, true); 
    while first < next  do
 
        # first catch a trivial case
        if next - first = 1  then
            Append( base, FactorArg( N, M ).generators );
            Add( firsts, next );
            Add( central, IsFCentralFactor( R, N, M ) );
            return rec( base := base,
                        firsts := firsts,
                        central := central );
        fi;

        # try to find a central factor
        U := FCommutatorSubgroup( R, N, M );
        index := Index( N, U );

        # if U is a proper subgroup we obtain a central factor
        if index > 1  then
            index := Length( Factors( index ) );
            Append( base, FactorArg( N, U ).generators );
            first := first + index;
            Add( firsts, first );
            Append( central, List( [ 1 .. index ], x -> true ) );
            N := Copy( U );

        # there is no central factor, check whether it is semisimple
        elif not wt[3] in List( G.weights{[1..head-1]}, x -> x[3]) then
            Append( base, FactorArg( N, M ).generators );
            Append( central, List( [ 1 .. next-first ], x -> false ) );
            Add( firsts, next );
            first := next;
       
        # now we need to compute a hypereccentric factor
        else

            # compute induced nilpotent series with large factors of R
            series := G.operations.InducedNilpotentSeries( G, R );

            # loop over it 
            h := 2;
            while index = 1  do
                U := FCommutatorSubgroup( series[ h ], N, M );
                index := Index( N, U );
                h := h + 1;
            od;   

            # now we have it
            index := Length( Factors( index ) );
            Append( base, FactorArg( N, U ).generators );
            Append( central, List( [ 1 .. index ], x -> false ) );
            first := first + index;
            Add( firsts, first );
            N := Copy( U );
        fi;
    od;
    return rec( base    := base, 
                firsts  := firsts, 
                central := central );
end;
    
#############################################################################
##
#F  SagGroupOps.FNormalizer( G, form ) . . . . . . . <form>-normalizer of <G>
##
SagGroupOps.FNormalizer := function( G, form )

    local max,              # length of cgs of G
          system,           # system record to store information
          primes,           # of the group
          locals,           # local residuals or their p-complements
          p, j,             # prime and position
          reflayer,         # refined layer
          head, tail, next, # indicate layers
          gens,             # generators of normalizer
          i;                # indices

    # if G is trivial return
    max := Length( G.generators );
    if max = 0 then return G; fi;

    # set up local residuals
    primes  := Set( List( G.weights, x -> x[3] ) );
    locals  := List( primes, x -> form.LocalResidual(G, x) );

    # set up system record with new base
    i := Position( G.first, G.head[2] );
    system         := rec();
    system.H       := G;
    system.base    := G.generators{ [ 1 .. G.head[ 2 ] - 1 ] };
    system.Flayers := G.layers{[ 1 .. G.head[ 2 ] - 1 ]};
    system.Ffirst  := G.first{ [ 1 .. i ] };
    if IsBound( form.support ) then
        system.Fcentral := List( [1..G.head[2]-1], 
                                  x -> G.weights[x][3] in form.support );
    else
        system.Fcentral := List( [1..G.head[2]-1], x -> true );
    fi;

    # refine elementary abelian series of G
    InfoForm2("#I refine series \n");
    while i < Length( G.first ) do
        p := G.weights[G.first[i]][3];

        if not IsBound( form.support ) or
           (IsBound( form.support ) and p in form.support)  then
            j := Position( primes, p );
            reflayer := G.operations.RefinedBaseLayer( G, locals[j], i );

            # append result
            Append( system.base, reflayer.base );
            Append( system.Ffirst, reflayer.firsts );
            Append( system.Fcentral, reflayer.central );
            i := i + 1;

            # if a head is completely central, then we add whole layer
            if G.first[i] in G.tail then
                j := Position( G.tail, G.first[i] );
                head  := G.head[j];
                tail  := G.tail[j];
                next  := G.head[j+1];
                if ForAll( system.Fcentral{[head..tail-1]}, x -> x ) 
                   and next > tail then
                    Append( system.base, G.generators{[tail..next-1]} );
                    Append( system.Fcentral, List([tail..next-1], x -> true));
                    j := Position( G.first, next );
                    Append( system.Ffirst, G.first{[i..j]});
                    i := j;
                fi;
            fi;
        else
            next   := G.first[ i+1 ];
            Append( system.base, G.generators{[G.first[i]..next-1]});
            Append( system.Ffirst, [next] );
            Append( system.Fcentral, List( [G.first[i]..next-1], x -> false ) );
            i := i + 1;
        fi;
    od;

    # compute depths
    system.depths := List( system.base, DepthAgWord );

    # compute layers
    system.Flayers := List( system.base, x -> false );
    j := 1;
    for i in [1..Length( system.base )] do
        if system.Ffirst[j] = i then
            j := j + 1;
        fi;
        system.Flayers[i] := j-1;
    od;

    # add weights
    system.weights := G.weights;

    # calculate p-complements of residuals
    for i in [1..Length(primes)] do
        if not IsBound( form.support ) or
           (IsBound( form.support) and primes[i] in form.support) then
            locals[i] := G.operations.PComplement(G, locals[i], primes[i]);
        fi;
    od;

    # run through series and change F-central factors
    gens := [ ];
    i := max;
    while i >= 1  and system.Fcentral[ i ]  do
        i := i - 1;
    od;
    while i >= 1  do
        if system.Fcentral[ i ]  then
            p := G.weights[ i ][ 3 ];
            j := Position( primes, p );
            G.operations.ChangeGenerator( system, locals[ j ], i );
        fi;
        i := i - 1;
    od;

    gens := system.base{ Filtered( [1..Length(system.Fcentral)], 
                         x -> system.Fcentral[x] ) };
    Sort( gens, function(x, y) return DepthAgWord(x) < DepthAgWord(y); end );
    return G.operations.AgSubgroup( G, gens, false );
end;

#############################################################################
##
#F  AgGroupOps.FNormalizer( G, form ) . . . . . . . .<form>-normalizer of <G>
##
AgGroupOps.FNormalizer := function( G, form )
    local S, N;
    if Length( G.generators ) = 0 then return G; fi;
    InfoForm2("#I compute special ag group \n");
    S := SpecialAgGroup( G );
    InfoForm2("#I compute normalizer \n");
    N := S.operations.FNormalizer( S, form );
    return AgSubgroup( G, List(Igs(N), x -> Image(S.bijection, x)), false );
end;

#############################################################################
##
#F  FNormalizer( G, form ) . . . . . . . . . . . . . <form>-normalizer of <G> 
##
FNormalizer := function( G, form )
    local func, str, i;

    # check argruments
    if not IsAgGroup( G ) or not IsFormation( form )  then
        Error(" Usage: FNormalizer( <G>, <formation> )\n");
    fi;
    if not IsBound( form.LocalResidual )  then
        Error(" Usage: <formation> needs to be locally defined \n");
    fi;
    if not IsBound( form.isIntegrated ) or not form.isIntegrated  then
        Error(" Usage: <formation> needs to be integrated \n");
    fi;

    # set up record structure 
    if not IsBound( G.formation ) then
        G.formation := rec( );
    fi;

    if not IsBound( form.support ) then

        if not IsBound( G.formation.(form.name) ) then 
            G.formation.(form.name) := rec( );
        fi;

        # check whether we know result
        if not IsBound( G.formation.(form.name).FNormalizer ) then

            # use the right function
            if IsBound( form.FNormalizer ) then
                G.formation.(form.name).FNormalizer := form.FNormalizer( G );
            else
                G.formation.(form.name).FNormalizer := 
                             G.operations.FNormalizer( G, form );
            fi;
        fi;

        # return
        return G.formation.(form.name).FNormalizer;
    else
        
        # turn support into string
        str := ConcatenationString( "Support", String( form.support[1] ) );
        for i in [2..Length( form.support )] do
            str := ConcatenationString( str, "_", String( form.support[i] ) );
        od;
        str := ConcatenationString( form.name, str );

        # add another record component
        if not IsBound( G.formation.(str) ) then
            G.formation.(str) := rec();
        fi;

        # check whether result is known
        if not IsBound( G.formation.(str).FNormalizer ) then

            # use the right function
            if IsBound( form.FNormalizer ) then
                G.formation.(str).FNormalizer := 
                                        form.FNormalizer( G );
            else
                G.formation.(str).FNormalizer := 
                             G.operations.FNormalizer( G, form );
            fi;
        fi;

        # return
        return G.formation.(str).FNormalizer;
    fi;
end;


