#############################################################################
##
#F  FResidual( <G>, <form> ) . . . . . . . . . . . . . <form>-residual of <G> 
##
FResidual := function( G, form )
    local func, str, i;

    # check argruments
    if not IsAgGroup( G ) or not IsFormation( form ) then
        Error(" Usage: FResidual( <G>, <formation> )\n");
    fi;

    # set up record structure in G
    if not IsBound( G.formation ) then
        G.formation := rec( );
    fi;

    if not IsBound( form.support ) then

        if not IsBound( G.formation.(form.name) ) then 
            G.formation.(form.name) := rec( );
        fi;

        # check whether we know result
        if not IsBound( G.formation.(form.name).FResidual ) then

            # use the right function
            if IsBound( form.FResidual ) then
                G.formation.(form.name).FResidual := form.FResidual( G );
            else
                G.formation.(form.name).FResidual := 
                                        G.operations.FResidual( G, form );
            fi;
        fi;

        # return
        return G.formation.(form.name).FResidual;
    else
        
        # turn support into string
        str := ConcatenationString( "Support", String( form.support[1] ) );
        for i in [2..Length( form.support )] do
            str := ConcatenationString( str, "_", String( form.support[i] ) );
        od;
        str := ConcatenationString( form.name, str );

        # add another record component
        if not IsBound( G.formation.(str) ) then
            G.formation.(str) := rec();
        fi;

        # check whether result is known
        if not IsBound( G.formation.(str).FResidual ) then

            # use the right function
            if IsBound( form.FResidual ) then
                G.formation.(str).FResidual := form.FResidual( G );
            else
                G.formation.(str).FResidual := 
                                        G.operations.FResidual( G, form );
            fi;
        fi;

        # return
        return G.formation.(str).FResidual;
    fi;
end;

#############################################################################
##
#F  AgGroupOps.FResidual( G, form ) . . . . . . . . . .<form>-residual in <G>
##
AgGroupOps.FResidual := function( G, form )

    local gens, gensR, # normal generators of residual 
          primes, sub, # lists of primes 
          p, q,        # primes
          N, M,        # approximate O^p'
          g,           # element
          U;           # subgroup

    Cgs( G );

    # get list of relevant primes
    primes := List( G.cgs, x -> RelativeOrderAgWord(x) );
    if IsBound( form.support ) then
        sub := Intersection( Set( primes ), form.support );
    else
        sub := Set( primes );
    fi;

    # set up gens with sub' part
    gens := G.cgs{Filtered( [1..Length(primes)], x -> not primes[x] in sub )};

    # loop over primes
    for p in sub do

        # compute local residual
        N := form.LocalResidual( G, p );

        # compute O^p
        M := N.operations.PResidual( N, p );

        # compute normal subgroup genset of O^p'
        gensR := [ ];
        for g  in Cgs( M ) do
            q := RelativeOrderAgWord( g );
            if q = p then
                Add( gensR, PPart( M, g, q ) );
            fi;
        od;

        # append it to the other generators
        Append( gens, gensR );
    od;

    # compute subgroup and normal closure
    U := G.operations.Subgroup( G, gens );
    return G.operations.NormalClosure( G, U );
end;

#############################################################################
##
#F  SagGroupOps.FResidual( G, form ) . . . . . . . . . <form>-residual in <G>
##
SagGroupOps.FResidual := AgGroupOps.FResidual;
